/*
 * Copyright (C) 2002-2022 CERN for the benefit of the ATLAS collaboration.
 */
/**
 * @file BatchedAlgorithm.cpp
 * @author scott snyder <snyder@bnl.gov>
 * @date Aug, 2022
 * @brief 
 */

#include <Gaudi/BatchedAlgorithm.h>
#include "GaudiKernel/FunctionalFilterDecision.h"
#include "GaudiKernel/Guards.h"


namespace Gaudi {


StatusCode BatchedAlgorithm::execute( const EventContext& ctx ) const
{
  // Called when we run in serial mode.
  return execute (std::vector<const EventContext*> { &ctx });
}


StatusCode BatchedAlgorithm::sysExecute( const EventContext& ctx )
{
  // Called when we run in serial mode.
  return Gaudi::Algorithm::sysExecute (ctx);
}


StatusCode BatchedAlgorithm::sysExecute( const std::vector<EventContext*>& ctxVec )
{
  // FIXME TODO: Reduce code duplication with Algorithm::sysExecute.
  
  if ( !isEnabled() ) {
    if ( msgLevel( MSG::VERBOSE ) ) { verbose() << ".sysExecute(): is not enabled. Skip execution" << endmsg; }
    return StatusCode::SUCCESS;
  }

  StatusCode status;

  std::vector<AlgExecState*> algStates;
  algStates.reserve (ctxVec.size());

  std::list<Gaudi::Utils::AlgContext> algContexts;
  IAlgContextSvc* ctxsvc = registerContext() ? contextSvc().get() : nullptr;

  std::list<Gaudi::Guards::AuditorGuard> auditorGuards;
  IAuditorSvc* audSvc = auditorExecute() ? auditorSvc().get() : nullptr;
  
  for (const EventContext* ctx: ctxVec) {
    algStates.push_back (&execState (*ctx) );
    algStates.back()->setState( AlgExecState::State::Executing );
    // lock the context service
    algContexts.emplace_back (this, ctxsvc, *ctx);
    if (audSvc) {
      auditorGuards.emplace_back (this, audSvc, IAuditor::Execute, status);
    }
  }

  // Should performance profile be performed ?
  // invoke execute() method of Algorithm class
  //   and catch all uncaught exceptions

  std::list<ITimelineSvc::TimelineRecorder> timelineRecorders;
  try {
    if (timelineSvc()->isEnabled()) {
      for (const EventContext* ctx: ctxVec) {
        timelineRecorders.emplace_back (timelineSvc()->getRecorder( name(), *ctx ) );
      }
    }

    status = execute( std::vector<const EventContext*> (ctxVec.begin(), ctxVec.end()) );

    if ( status == Gaudi::Functional::FilterDecision::FAILED ) {
      // TODO: Filter failure for one fails the entire batch.
      for (AlgExecState* state : algStates) {
        state->setFilterPassed( false );
      }
    } else if ( status.isFailure() ) {
      status = exceptionSvc()->handleErr( *this, status );
    }

  } catch ( const GaudiException& Exception ) {

    if ( Exception.code() == StatusCode::FAILURE ) {
      fatal();
    } else {
      error() << " Recoverable";
    }

    msgStream() << " Exception with tag=" << Exception.tag() << " is caught " << endmsg;

    error() << Exception << endmsg;

    // Stat stat( chronoSvc() , Exception.tag() ) ;
    status = exceptionSvc()->handle( *this, Exception );
  } catch ( const std::exception& Exception ) {

    fatal() << " Standard std::exception is caught " << endmsg;
    error() << Exception.what() << endmsg;
    // Stat stat( chronoSvc() , "*std::exception*" ) ;
    status = exceptionSvc()->handle( *this, Exception );
  } catch ( ... ) {

    fatal() << "UNKNOWN Exception is caught " << endmsg;
    // Stat stat( chronoSvc() , "*UNKNOWN Exception*" ) ;

    status = exceptionSvc()->handle( *this );
  }

  if ( status.isFailure() ) {
    // Increment the error count
    algExecStateSvc()->incrementErrorCount( this );
    error() << "Error executing algorithm." << endmsg;
  }

  for (AlgExecState* state : algStates) {
    state->setState( AlgExecState::State::Done, status );
  }

  return status;
}


} // namespace Gaudi

